"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.isScopedPackagePath = exports.doResolvePath = exports.doResolve = void 0;
var tslib_1 = require("tslib");
var resolve_1 = (0, tslib_1.__importDefault)(require("resolve"));
var fs_1 = (0, tslib_1.__importDefault)(require("fs"));
var path_1 = (0, tslib_1.__importDefault)(require("path"));
/**
 * An attempt at drying out filesystem references to [external packages](https://tinyurl.com/mao2dy6).
 *
 * We access external packages for a number of reasons:
 *  - to simply resolve a non-scoped package's (absolute) path on disk. e.g. '/var/project/node_modules/react-native'
 *  - to resolve a scoped package's (absolute) path on disk. e.g. '/var/project/node_modules/@aScope/react-native'
 *  - to resolve an individual file/dir's (absolute) path on disk. e.g. '/var/project/node_modules/react-native/blur/android' (note this is a normally illegal package reference)
 *  - to resolve an individual file (absolute) path on disk by FS relative reference. e.g. '/var/project/node_modules/react-native/relPath'
 *
 * ** Please note that we do not support [subpackage paths](https://tinyurl.com/vub6c7t). All suffixed paths (e.g. 'react-native/SUFFIX_PATH', '@aScope/react-native/SUFFIX_PATH')
 * will be treated as a filepath from root of resolved package (i.e. will ignore subdirectory package.json)
 *
 * @param {*} aPath - package name. e.g. 'file:../rel/path', 'react-native', 'react-native/android', '@react-native-community/masked-view/android'
 * @param {*} mandatory - whether it throws
 * @param {*} options - docs - https://tinyurl.com/r9sfpf7 && {keepSuffix: boolean}
 */
var doResolve = function (aPath, mandatory, options) {
    var _a;
    if (mandatory === void 0) { mandatory = true; }
    if (options === void 0) { options = {}; }
    options.basedir = (_a = options.basedir) !== null && _a !== void 0 ? _a : process.cwd();
    try {
        if (aPath.startsWith('file:')) {
            return _withPathFix(_doResolveFSPath(aPath, options), options === null || options === void 0 ? void 0 : options.forceForwardPaths);
        }
        return _withPathFix(_doResolveExternalPackage(aPath, options), options === null || options === void 0 ? void 0 : options.forceForwardPaths);
    }
    catch (err) {
        // perhaps do some warning logging here..
        if (mandatory)
            throw err;
    }
};
exports.doResolve = doResolve;
var _withPathFix = function (p, forceForwardPaths) {
    if (p && forceForwardPaths) {
        return p.replace(/\\/g, '/');
    }
    return p;
};
var doResolvePath = function (aPath, mandatory, options, fallbackBase) {
    var _a;
    if (mandatory === void 0) { mandatory = true; }
    if (options === void 0) { options = {}; }
    if (fallbackBase === void 0) { fallbackBase = ''; }
    options.basedir = (_a = options.basedir) !== null && _a !== void 0 ? _a : process.cwd();
    try {
        var pathArr = aPath.split('/');
        // Take care of scenario when someone wrote: "/node_modules/.." instead of "node_modules/..."
        if (pathArr[0] === '') {
            pathArr.shift();
        }
        if (pathArr[0] === 'node_modules') {
            pathArr.shift();
        }
        if (pathArr[0] === 'packages') {
            pathArr.shift();
        }
        var cleanPath = pathArr.join('/');
        if (pathArr[0].startsWith('@')) {
            pathArr.shift();
        }
        pathArr.shift();
        var realPath = (0, exports.doResolve)(cleanPath, mandatory, options);
        if (realPath) {
            return _withPathFix(path_1.default.join.apply(path_1.default, (0, tslib_1.__spreadArray)([realPath], pathArr, false)), options === null || options === void 0 ? void 0 : options.forceForwardPaths);
        }
        return _withPathFix(path_1.default.join(fallbackBase, aPath), options === null || options === void 0 ? void 0 : options.forceForwardPaths);
    }
    catch (err) {
        if (mandatory)
            throw err;
    }
};
exports.doResolvePath = doResolvePath;
var isScopedPackagePath = function (aPath) {
    if (aPath.startsWith('@')) {
        if (!aPath.includes('/')) {
            throw new Error("Scoped packages must include subpackage portion e.g. '@aScope/subpackage'. Supplied path: ".concat(aPath));
        }
        return true;
    }
};
exports.isScopedPackagePath = isScopedPackagePath;
var _getPackagePathParts = function (aPath) {
    var parts = [];
    if ((0, exports.isScopedPackagePath)(aPath)) {
        parts = aPath.match(/^([^/]+\/[^/]+)(?:\/?(.*))/);
    }
    else {
        parts = aPath.match(/^([^/]+)\/?(.*)/);
    }
    if (!Array.isArray(parts)) {
        throw new Error("Unsuitable path for resolving: ".concat(aPath));
    }
    return parts.slice(1);
};
/**
 * We support path linking using 'file:' protocol (not part of official node resolution alg.)
 */
var _doResolveFSPath = function (aPath, options) {
    var fileRelPath = "".concat(options.basedir ? "".concat(options.basedir, "/").replace(/.*\/+$/, '/') : '').concat(aPath.replace('file:', ''));
    if (!fs_1.default.existsSync(fileRelPath)) {
        throw new Error("Explicit filepath ".concat(aPath, " does not resolve to dir or file"));
    }
    return fileRelPath;
};
/**
 * @see 'LOAD_NODE_MODULES' of node resolution alg. - https://tinyurl.com/pgz6f33
 */
var _doResolveExternalPackage = function (aPath, options) {
    var _a, _b;
    var _c = _getPackagePathParts(aPath), packageBase = _c[0], packageSuffix = _c[1];
    try {
        var resolvedPath = resolve_1.default
            .sync(packageBase, (0, tslib_1.__assign)((0, tslib_1.__assign)({ packageFilter: function (pkg) {
                pkg.main = 'package.json';
                return pkg;
            } }, options), { extensions: ['.js', '.json'].concat((_a = options.extensions) !== null && _a !== void 0 ? _a : []) }))
            .replace(/(\\|\/)package.json$/, '');
        return ((_b = options.keepSuffix) !== null && _b !== void 0 ? _b : false)
            ? "".concat(resolvedPath, "/").concat(packageSuffix)
            : resolvedPath;
    }
    catch (e) {
        return null;
    }
};
//# sourceMappingURL=resolve.js.map